from ..models.audio_encoding import AudioEncoding
import queue
from typing import Optional

class BaseInputDevice():

    def __init__(self, sampling_rate: int, audio_encoding: AudioEncoding, chunk_size: int):
        self.sampling_rate = sampling_rate
        self.audio_encoding = audio_encoding
        self.chunk_size = chunk_size
        self.queue = queue.Queue()

    def get_audio(self) -> Optional[bytes]:
        raise NotImplementedError


import sounddevice as sd
import numpy as np
from typing import Optional
import queue
import wave

from .base_input_device import BaseInputDevice
from ..models.audio_encoding import AudioEncoding

class MicrophoneInput(BaseInputDevice):

    DEFAULT_SAMPLING_RATE = 44100
    DEFAULT_CHUNK_SIZE = 2048

    def __init__(self, device_info: dict, sampling_rate: int = None, chunk_size: int = DEFAULT_CHUNK_SIZE, microphone_gain: int = 1):
        self.device_info = device_info
        sampling_rate = sampling_rate or (self.device_info.get('default_samplerate', self.DEFAULT_SAMPLING_RATE))
        super().__init__(int(sampling_rate), AudioEncoding.LINEAR16, chunk_size)
        self.stream = sd.InputStream(
            dtype=np.int16,
            channels=1,
             samplerate=self.sampling_rate,
            blocksize=self.chunk_size,
            device=int(self.device_info['index']),
            callback=self._stream_callback
        )
        self.stream.start()
        self.queue = queue.Queue()
        self.microphone_gain = microphone_gain

    def _stream_callback(self, in_data: np.ndarray[np.int16], *_args):
        if self.microphone_gain > 1:
            in_data = in_data * (2 ^ self.microphone_gain)
        else:
            in_data = in_data // (2 ^ self.microphone_gain)
        audio_bytes = in_data.tobytes()
        self.queue.put_nowait(audio_bytes)

    def get_audio(self) -> Optional[bytes]:
        try:
            return self.queue.get_nowait()
        except queue.Empty:
            return None


from enum import Enum
from typing import Optional
from .model import TypedModel
from .audio_encoding import AudioEncoding
from ..output_device.base_output_device import BaseOutputDevice

class SynthesizerType(str, Enum):
    BASE = "synthesizer_base"
    AZURE = "synthesizer_azure"
    GOOGLE = "synthesizer_google"
    ELEVEN_LABS = "synthesizer_eleven_labs"

class SynthesizerConfig(TypedModel, type=SynthesizerType.BASE):
    sampling_rate: int
    audio_encoding: AudioEncoding

    @classmethod
    def from_output_device(cls, output_device: BaseOutputDevice):
        return cls(sampling_rate=output_device.sampling_rate, audio_encoding=output_device.audio_encoding)

class AzureSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.AZURE):
    voice_name: Optional[str] = None
    pitch: Optional[int] = None
    rate: Optional[int] = None

    @classmethod
    def from_output_device(
        cls, 
        output_device: BaseOutputDevice, 
        voice_name: Optional[str] = None,
        pitch: Optional[int] = None,
        rate: Optional[int] = None,
    ):
        return cls(
            sampling_rate=output_device.sampling_rate, 
            audio_encoding=output_device.audio_encoding,
            voice_name=voice_name,
            pitch=pitch,
            rate=rate,
        )
    pass

class GoogleSynthesizerConfig(SynthesizerConfig, type=SynthesizerType.GOOGLE):
    pass



from fastapi import FastAPI, Response, Form
import requests
import uvicorn
from .. import api_key, BASE_URL

from ..models.agent import AgentConfig
from ..models.telephony import CreateInboundCall

VOCODE_INBOUND_CALL_URL = f"https://{BASE_URL}/create_inbound_call"

class InboundCallServer():

    def __init__(self, agent_config: AgentConfig):
        self.agent_config = agent_config
        self.app = FastAPI()
        self.app.post("/vocode")(self.handle_call)

    async def handle_call(self, twilio_sid: str = Form(alias='CallSid')):
        response = requests.post(
            VOCODE_INBOUND_CALL_URL,
            headers={
                "Authorization": f"Bearer {api_key}"
            },
            json=CreateInboundCall(
                agent_config=self.agent_config,
                twilio_sid=twilio_sid
            ).dict()
        )
        assert response.ok
        return Response(
            response.text,
            media_type="application/xml",
        )
    
    def run(self, host="localhost", port=3000):
        uvicorn.run(self.app, host=host, port=port)


import pydantic

class BaseModel(pydantic.BaseModel):

    def __init__(self, **data):
        for key, value in data.items():
            if isinstance(value, dict):
                if 'type' in value:
                    data[key] = TypedModel.parse_obj(value)
        super().__init__(**data)

# Adapted from https://github.com/pydantic/pydantic/discussions/3091
class TypedModel(BaseModel):

    _subtypes_ = []

    def __init_subclass__(cls, type=None):
        cls._subtypes_.append([type, cls])

    @classmethod
    def get_cls(_cls, type):
        for t, cls in _cls._subtypes_:
            if t == type:
                return cls
        raise ValueError(f'Unknown type {type}')
    
    @classmethod
    def get_type(_cls, cls_name):
        for t, cls in _cls._subtypes_:
            if cls.__name__ == cls_name:
                return t
        raise ValueError(f'Unknown class {cls_name}')
    
    @classmethod
    def parse_obj(cls, obj):
        data_type = obj.get('type')
        if data_type is None:
            raise ValueError(f'type is required for {cls.__name__}')
    
        sub = cls.get_cls(data_type)
        if sub is None:
            raise ValueError(f'Unknown type {data_type}')
        return sub(**obj)

    def _iter(self, **kwargs):
        yield 'type', self.get_type(self.__class__.__name__)
        yield from super()._iter(**kwargs)

    @property
    def type(self):
        return self.get_type(self.__class__.__name__)




from typing import Optional
from vocode.models.model import BaseModel
from vocode.models.agent import AgentConfig
from vocode.models.synthesizer import SynthesizerConfig


class CallEntity(BaseModel):
    phone_number: str


class CreateInboundCall(BaseModel):
    agent_config: AgentConfig
    twilio_sid: str


class CreateOutboundCall(BaseModel):
    recipient: CallEntity
    caller: CallEntity
    agent_config: AgentConfig
    synthesizer_config: Optional[SynthesizerConfig] = None
    # TODO add IVR/etc.



from typing import Optional
from enum import Enum
from .model import TypedModel, BaseModel


class AgentType(str, Enum):
    BASE = "agent_base"
    LLM = "agent_llm"
    CHAT_GPT_ALPHA = "agent_chat_gpt_alpha"
    CHAT_GPT = "agent_chat_gpt"
    ECHO = "agent_echo"
    INFORMATION_RETRIEVAL = "agent_information_retrieval"
    RESTFUL_USER_IMPLEMENTED = "agent_restful_user_implemented"
    WEBSOCKET_USER_IMPLEMENTED = "agent_websocket_user_implemented"


class AgentConfig(TypedModel, type=AgentType.BASE):
    initial_message: Optional[str] = None
    generate_responses: bool = True
    allowed_idle_time_seconds: Optional[float] = None

class LLMAgentConfig(AgentConfig, type=AgentType.LLM):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None

class ChatGPTAlphaAgentConfig(AgentConfig, type=AgentType.CHAT_GPT_ALPHA):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None

class ChatGPTAgentConfig(AgentConfig, type=AgentType.CHAT_GPT):
    prompt_preamble: str
    expected_first_prompt: Optional[str] = None
    generate_responses: bool = False

class InformationRetrievalAgentConfig(
    AgentConfig, type=AgentType.INFORMATION_RETRIEVAL
):
    recipient_descriptor: str
    caller_descriptor: str
    goal_description: str
    fields: list[str]
    # TODO: add fields for IVR, voicemail


class EchoAgentConfig(AgentConfig, type=AgentType.ECHO):
    pass

class RESTfulUserImplementedAgentConfig(AgentConfig, type=AgentType.RESTFUL_USER_IMPLEMENTED):
    class EndpointConfig(BaseModel):
        url: str
        method: str = "POST"

    respond: EndpointConfig
    generate_responses: bool = False
    # generate_response: Optional[EndpointConfig]
    # update_last_bot_message_on_cut_off: Optional[EndpointConfig]

class RESTfulAgentInput(BaseModel):
    human_input: str

class RESTfulAgentOutputType(str, Enum):
    BASE = "restful_agent_base"
    TEXT = "restful_agent_text"
    END = "restful_agent_end"

class RESTfulAgentOutput(TypedModel, type=RESTfulAgentOutputType.BASE):
    pass

class RESTfulAgentText(RESTfulAgentOutput, type=RESTfulAgentOutputType.TEXT):
    response: str

class RESTfulAgentEnd(RESTfulAgentOutput, type=RESTfulAgentOutputType.END):
    pass

class WebSocketUserImplementedAgentConfig(AgentConfig, type=AgentType.WEBSOCKET_USER_IMPLEMENTED):
    class RouteConfig(BaseModel):
        url: str

    respond: RouteConfig
    generate_responses: bool = False
    # generate_response: Optional[RouteConfig]
    # send_message_on_cut_off: bool = False

class WebSocketAgentMessageType(str, Enum):
    BASE = 'websocket_agent_base'
    START = 'websocket_agent_start'
    TEXT = 'websocket_agent_text'
    READY = 'websocket_agent_ready'
    STOP = 'websocket_agent_stop'

class WebSocketAgentMessage(TypedModel, type=WebSocketAgentMessageType.BASE): pass

class WebSocketAgentTextMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.TEXT):
    class Payload(BaseModel):
        text: str

    data: Payload

    @classmethod
    def from_text(cls, text: str):
        return cls(data=cls.Payload(text=text))


class WebSocketAgentStartMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.START):
    pass

class WebSocketAgentReadyMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.READY):
    pass

class WebSocketAgentStopMessage(WebSocketAgentMessage, type=WebSocketAgentMessageType.STOP):
    pass



from ..models.telephony import CallEntity, CreateOutboundCall
import requests
from .. import api_key, BASE_URL

VOCODE_OUTBOUND_CALL_URL = f"https://{BASE_URL}/create_outbound_call"

class OutboundCall:

    def __init__(self, recipient: CallEntity, caller: CallEntity, agent_config):
        self.recipient = recipient
        self.caller = caller
        self.agent_config = agent_config

    def start(self):
        return requests.post(
            VOCODE_OUTBOUND_CALL_URL,
            headers={
                "Authorization": f"Bearer {api_key}"
            },
            json=CreateOutboundCall(
                recipient=self.recipient,
                caller=self.caller,
                agent_config=self.agent_config
            ).dict()
        )

    


import sounddevice as sd
from .input_device.microphone_input import MicrophoneInput
from .output_device.speaker_output import SpeakerOutput
import logging

logger = logging.getLogger(__name__)

def _get_device_prompt(device_infos: list[dict]) -> str:
    return """Please select a device:
{}
Choice: """.format(
        "\n".join(f"{index}: {device['name']}" for index, device in enumerate(device_infos)))

def create_microphone_input_and_speaker_output(use_default_devices=False, mic_sampling_rate=None, speaker_sampling_rate=None) -> tuple[MicrophoneInput, SpeakerOutput]:
    device_infos = sd.query_devices()
    input_device_infos = list(filter(lambda device_info: device_info['max_input_channels'] > 0, device_infos))
    output_device_infos = list(filter(lambda device_info: device_info['max_output_channels'] > 0, device_infos))
    if use_default_devices:
        input_device_info = sd.query_devices(kind='input')
        output_device_info = sd.query_devices(kind='output')
    else:
        input_device_info = input_device_infos[int(input(_get_device_prompt(input_device_infos)))]
        output_device_info = output_device_infos[int(input(_get_device_prompt(output_device_infos)))]
    logger.info("Using microphone input device: %s", input_device_info['name'])
    microphone_input = MicrophoneInput(input_device_info, sampling_rate=mic_sampling_rate)
    logger.info("Using speaker output device: %s", output_device_info['name'])
    speaker_output = SpeakerOutput(output_device_info, sampling_rate=speaker_sampling_rate) 
    return microphone_input, speaker_output


from fastapi import FastAPI
import uvicorn

class BaseAgent():

    def __init__(self):
        self.app = FastAPI()
    
    def run(self, host="localhost", port=3000):
        uvicorn.run(self.app, host=host, port=port)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vocode',
 'vocode.input_device',
 'vocode.models',
 'vocode.output_device',
 'vocode.telephony',
 'vocode.user_implemented_agent']

package_data = \
{'': ['*']}

install_requires = \
['anyio==3.6.2',
 'certifi==2022.12.7',
 'cffi==1.15.1',
 'charset-normalizer==3.0.1',
 'click==8.1.3',
 'decorator==5.1.1',
 'fastapi==0.92.0',
 'h11==0.14.0',
 'idna==3.4',
 'numpy==1.24.2',
 'ply==3.11',
 'pycparser==2.21',
 'pydantic==1.10.5',
 'python-dotenv==0.21.1',
 'requests==2.28.2',
 'six==1.16.0',
 'sniffio==1.3.0',
 'sounddevice==0.4.6',
 'starlette==0.25.0',
 'typing-extensions==4.5.0',
 'urllib3==1.26.14',
 'uvicorn==0.20.0',
 'websockets==10.4']

setup_kwargs = {
    'name': 'vocode',
    'version': '0.1.25',
    'description': 'The all-in-one voice SDK',
    'long_description': '# vocode Python SDK\n\n```\npip install vocode\n```\n\n```python\nimport asyncio\nimport signal\n\nfrom vocode.conversation import Conversation\nfrom vocode.helpers import create_microphone_input_and_speaker_output\nfrom vocode.models.transcriber import DeepgramTranscriberConfig\nfrom vocode.models.agent import LLMAgentConfig\nfrom vocode.models.synthesizer import AzureSynthesizerConfig\n\nif __name__ == "__main__":\n    microphone_input, speaker_output = create_microphone_input_and_speaker_output(use_first_available_device=True)\n\n    conversation = Conversation(\n        input_device=microphone_input,\n        output_device=speaker_output,\n        transcriber_config=DeepgramTranscriberConfig.from_input_device(microphone_input),\n        agent_config=LLMAgentConfig(prompt_preamble="The AI is having a pleasant conversation about life."),\n        synthesizer_config=AzureSynthesizerConfig.from_output_device(speaker_output)\n    )\n    signal.signal(signal.SIGINT, lambda _0, _1: conversation.deactivate())\n    asyncio.run(conversation.start())\n```\n',
    'author': 'Ajay Raj',
    'author_email': 'ajay@vocode.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .base_agent import BaseAgent
from ..models.agent import RESTfulAgentInput, RESTfulAgentOutput, RESTfulAgentText, RESTfulAgentEnd
from pydantic import BaseModel
from typing import Union
from fastapi import APIRouter

class RESTfulAgent(BaseAgent):
        
    def __init__(self):
        super().__init__()
        self.app.post("/respond")(self.respond_rest)

    async def respond(self, human_input) -> RESTfulAgentOutput:
        raise NotImplementedError

    async def respond_rest(self, request: RESTfulAgentInput) -> Union[RESTfulAgentText, RESTfulAgentEnd]:
        response = await self.respond(request.human_input)
        return response




from .base_agent import BaseAgent
import uuid
import typing
from typing import Union, Optional
from fastapi import WebSocket
from ..models.agent import (
    WebSocketAgentStartMessage, 
    WebSocketAgentReadyMessage, 
    WebSocketAgentTextMessage, 
    WebSocketAgentStopMessage, 
    WebSocketAgentMessage, 
    WebSocketAgentMessageType
)

class WebSocketAgent(BaseAgent):
        
    def __init__(self):
        super().__init__()
        self.app.websocket("/respond")(self.respond_websocket)

    async def respond(self, human_input: str, conversation_id: Optional[str] = None) -> Union[WebSocketAgentTextMessage, WebSocketAgentStopMessage]:
        raise NotImplementedError

    async def respond_websocket(self, websocket: WebSocket):
        await websocket.accept()
        conversation_id = str(uuid.uuid4())
        WebSocketAgentStartMessage.parse_obj(await websocket.receive_json())
        await websocket.send_text(WebSocketAgentReadyMessage().json())
        while True:
            input_message = WebSocketAgentMessage.parse_obj(await websocket.receive_json())
            if input_message.type == WebSocketAgentMessageType.STOP:
                break
            text_message = typing.cast(WebSocketAgentTextMessage, input_message)
            output_response = await self.respond(text_message.data.text, conversation_id=conversation_id)
            await websocket.send_text(output_response.json())
        await websocket.close()

